/*
 * ngx_http_oceanbase_handler.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#include "ngx_http_oceanbase_module.h"
#include "ngx_http_oceanbase_handler.h"
#include "ngx_http_oceanbase_processor.h"
#include "ngx_http_oceanbase_util.h"
#include "ngx_http_oceanbase_upstream.h"
#include "ngx_http_oceanbase_keepalive.h"
#include "ngx_tpcc_common.h"
#include "ngx_tpcc_txn.h"
#include "ngx_tpcc_delivery_pool.h"

static ngx_int_t ngx_http_oceanbase_create_request(ngx_http_request_t *r);
static ngx_int_t ngx_http_oceanbase_reinit_request(ngx_http_request_t *r);
static void ngx_http_oceanbase_abort_request(ngx_http_request_t *r);
static void ngx_http_oceanbase_finalize_request(ngx_http_request_t *r, ngx_int_t rc);
static ngx_int_t ngx_http_oceanbase_process_header(ngx_http_request_t *r);
static ngx_int_t ngx_http_oceanbase_input_filter_init(void *data);
static ngx_int_t ngx_http_oceanbase_input_filter(void *data, ssize_t bytes);
static void ngx_http_oceanbase_handler_inner(ngx_http_request_t *r);

ngx_thread_pool_t * g_del_thread_pool = NULL;

bool parse_get_method_params(u_char *params, int *t_id, int *w_id, int *sl_d_id)
{
	bool ret = false;
	bool bnull = false;

	if(params){
		if(*params++ == '?'){
			if(ngx_strncmp(params, "t_id=", 5) ==0){
				params += 5;
            	parse_param(params, *t_id, bnull);
				
				if(!bnull && ngx_strncmp(params, "&sl_d_id=", 9) ==0){
					params += 9;
                	parse_param(params, *sl_d_id, bnull);
					
					if(!bnull && ngx_strncmp(params, "&w_id=", 6) ==0){
                    	params += 6;
                    	parse_param(params, *w_id, bnull);

						if(!bnull && (*params == '\0' || *params == ' ')){
							*params = '\0';
							ret = true;
						}else{}
					}else{}
				}else{}
			}else{}
		}else{}
	}else{}

	return ret;
}

#define get_part_len(data, data_len, partlen)	do{\
											partlen = 0;\
											while(partlen <= data_len){\
												if(*(data + partlen) != '?'){\
													partlen ++;\
												}else{break;}\
											}\
										}while(0)


static ngx_int_t
ngx_http_tpcc_input_handler(ngx_http_request_t *r)
{
    ngx_int_t rc;
    int txn_type = TXN_UNKNOWN;
    char out_page[4096] = {'\0'};
    ngx_buf_t   *b;
    ngx_chain_t output;
    ngx_log_t * log = r->connection->log;
	int t_id = -1, w_id = -1, sl_d_id = -1;
	unsigned int partlen = 0;

    ngx_log_error(NGX_LOG_DEBUG, log, 0, "uri: %s %d\n", r->uri.data, r->uri.len);
	get_part_len(r->uri.data, r->uri.len, partlen);

    switch(partlen){
    case    sizeof(NEW_ORDER_INPUT_URI) - 1: //new-order-input
        if(!ngx_strncmp(r->uri.data, NEW_ORDER_INPUT_URI, sizeof(NEW_ORDER_INPUT_URI)-1)
			&& parse_get_method_params(r->uri.data + partlen, &t_id, &w_id, &sl_d_id)){
            sprintf(out_page, no_input_page, t_id, sl_d_id, w_id, r->uri.data + partlen);
            txn_type = NEW_ORDER;
        }else{}
        break;
    case    sizeof(PAYMENT_INPUT_URI) - 1:  //payment-input
        if(!ngx_strncmp(r->uri.data, PAYMENT_INPUT_URI, sizeof(PAYMENT_INPUT_URI) - 1)
			&& parse_get_method_params(r->uri.data + partlen, &t_id, &w_id, &sl_d_id)){
            sprintf(out_page, pay_input_page, t_id, sl_d_id, w_id, r->uri.data + partlen);
            txn_type = PAYMENT;
        }else{}
        break;
    case    sizeof(DELIVERY_INPUT_URI) - 1:  //delivery-input
        if(!ngx_strncmp(r->uri.data, DELIVERY_INPUT_URI, sizeof(DELIVERY_INPUT_URI) - 1)
			&& parse_get_method_params(r->uri.data + partlen, &t_id, &w_id, &sl_d_id)){
            sprintf(out_page, del_input_page, t_id, sl_d_id, w_id, r->uri.data + partlen);
            txn_type = DELIVERY;
        }else{}
        break;
    case    sizeof(ORDER_STATUS_INPUT_URI) - 1: //order-status-input
        if(!ngx_strncmp(r->uri.data, ORDER_STATUS_INPUT_URI, sizeof(ORDER_STATUS_INPUT_URI) - 1)
			&& parse_get_method_params(r->uri.data + partlen, &t_id, &w_id, &sl_d_id)){
            sprintf(out_page, os_input_page, t_id, sl_d_id, w_id, r->uri.data + partlen);
            txn_type = ORDER_STATUS;
        }else{}
        break;
    case    sizeof(STOCK_LEVEL_INPUT_URI) - 1: //stock-level-input
        if(!ngx_strncmp(r->uri.data, STOCK_LEVEL_INPUT_URI, sizeof(STOCK_LEVEL_INPUT_URI) - 1)
			&& parse_get_method_params(r->uri.data + partlen, &t_id, &w_id, &sl_d_id)){
            sprintf(out_page, sl_input_page, t_id, sl_d_id, w_id, sl_d_id, r->uri.data + partlen);
            txn_type = STOCK_LEVEL;
        }else{}
        break;
    case    sizeof(TXN_MENU_URI) - 1://txn-menu
        if(!ngx_strncmp(r->uri.data, TXN_MENU_URI, sizeof(TXN_MENU_URI) - 1)
			&& parse_get_method_params(r->uri.data + partlen, &t_id, &w_id, &sl_d_id)){
			char *hidden_params = (char *)(r->uri.data) + partlen;
            sprintf(out_page, menu_page, hidden_params, hidden_params, hidden_params, hidden_params, hidden_params);
            txn_type = TXN_MENU;
        }else{}
        break;
    default:
        txn_type =  TXN_UNKNOWN;
        break;
    }

    // default GET handler
    if( txn_type == TXN_UNKNOWN ){
		strcpy(out_page, logon_page);
    }

    rc = ngx_http_discard_request_body(r);
    if (rc != NGX_OK) {
        return rc;
    }

    r->headers_out.status = NGX_HTTP_OK;
    r->headers_out.content_length_n = strlen(out_page);
    ngx_str_set(&r->headers_out.content_type, "text/html");

    if (ngx_http_set_etag(r) != NGX_OK) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }

    if (ngx_http_set_content_type(r) != NGX_OK) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }

    r->allow_ranges = 1;
    b = ngx_pcalloc(r->pool, sizeof(ngx_buf_t));
    if (b == NULL) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }

    rc = ngx_http_send_header(r);

    if (rc == NGX_ERROR || rc > NGX_OK || r->header_only) {
        return rc;
    }

    // adjust the pointers of the buffer
    b->pos  = (u_char *)out_page;
    b->last = (u_char *)out_page + ngx_strlen(out_page);
    b->memory = 1;    // this buffer is in memory
    b->last_buf = 1;  // this is the last buffer in the buffer chain

    output.buf = b;
    output.next = NULL;

    return ngx_http_output_filter(r,&output);
}

ngx_int_t
ngx_http_oceanbase_handler(ngx_http_request_t *r)
{
    if (!(r->method & (NGX_HTTP_POST | NGX_HTTP_GET) ) ){
        return NGX_DECLINED;
    }

    if( r->method & NGX_HTTP_GET ){
        return ngx_http_tpcc_input_handler(r);
    }else{
        ngx_int_t rc = ngx_http_read_client_request_body(r, ngx_http_oceanbase_handler_inner);
        if (rc == NGX_ERROR || rc >= NGX_HTTP_SPECIAL_RESPONSE) {
            return rc;
        }
    }
    return NGX_AGAIN;
}

extern int process_delivery(u_char *params, char *out, int len, ngx_log_t *log);

static void ngx_http_oceanbase_handle_logon(ngx_http_request_t *r)
{
	ngx_chain_t *bufs = r->request_body->bufs;
	int len = 0;
	int ret = 0;
	u_char params_[128] = {'\0'};
	u_char *params = params_;
	char * out = NULL;
	ngx_buf_t *response_body;
	int	w_id = -1, sl_d_id = -1;
	bool bnull = false;

	while(bufs){
		 ngx_buf_t *buf = bufs->buf;
		 memcpy(params + len, buf->pos, buf->last - buf->pos);
		 len += (buf->last - buf->pos);
		 bufs = bufs->next;
	}

	params[len] = '\0';
	
	if(ngx_strncmp(params, "w_id=", 5) !=0 ){
		ret = -1;
	}else{
		params += 5;
		parse_param(params, w_id, bnull);
		if(bnull || (ngx_strncmp(params, "&sl_d_id=", 9) != 0)){
			ret =  -1;
		}else{
			params += 9;
			parse_param(params, sl_d_id, bnull);
			if(bnull){
				ret = -1;
			}else{}
		}
	}
		
	out = ngx_pcalloc(r->pool, 1024);
	if(out == NULL){
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                        "ngx_pcalloc failed:%d", NGX_HTTP_INTERNAL_SERVER_ERROR);
        return ; // NGX_HTTP_INTERNAL_SERVER_ERROR;
	}
	r->headers_out.status = NGX_HTTP_OK;
	ngx_str_set(&r->headers_out.content_type, "text/html");

	// allocate a buffer for your response body
	response_body = ngx_pcalloc(r->pool, sizeof(ngx_buf_t));
	if (response_body == NULL) {
		ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, 
						"ngx_pcalloc failed:%d", NGX_HTTP_INTERNAL_SERVER_ERROR);
		return ; // NGX_HTTP_INTERNAL_SERVER_ERROR;
	}

	if(ret == -1 || w_id <=0 || sl_d_id <= 0){
		char msg[128] = {0};
		sprintf(msg, "invalid parameters, w_id and d_id should be positive integer");
		sprintf(out, error_info_page, msg);	
	}else{
		char hidden[128] = {0}; //hidden_sl[128] = {0};
		sprintf(hidden, "?t_id=%d&sl_d_id=%d&w_id=%d", 10*(w_id - 1) + sl_d_id, sl_d_id, w_id);
		sprintf(out, menu_page, hidden, hidden, hidden, hidden, hidden);
	}

	// send response
	r->headers_out.content_length_n = ngx_strlen(out);
	int rc = ngx_http_send_header(r);
	if (rc == NGX_ERROR || rc > NGX_OK || r->header_only)
	{
		ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, "ngx_http_send_header err:%d", rc);
		return;
	}

	// adjust the pointers of the buffer
	response_body->pos = (u_char *)out;
	response_body->last = (u_char *)out + ngx_strlen(out);
	response_body->memory = 1;    // this buffer is in memory
	response_body->last_buf = 1;  // this is the last buffer in the buffer chain

	ngx_chain_t output;
	output.buf = response_body;
	output.next = NULL;

	r->main->count--;

	ngx_http_output_filter(r,&output);
	
	return;
}

static void ngx_http_oceanbase_handle_delivery(ngx_http_request_t *r, ngx_oceanbase_delivery_queue_t *dq)
{
	ngx_chain_t *bufs = r->request_body->bufs;
	int len = 0;
	int ret = 0;
	u_char params[128] = {'\0'};
	char * out = NULL;
	ngx_buf_t *response_body;
	ngx_str_t name = ngx_string(DELIVERY_THREAD_POOL);

	while(bufs){
		 ngx_buf_t *buf = bufs->buf;
		 memcpy(params + len, buf->pos, buf->last - buf->pos);
		 len += (buf->last - buf->pos);
		 bufs = bufs->next;
	}

	params[len] = '\0';
	
	out = ngx_pcalloc(r->pool, 1024);
	if(out == NULL){
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                        "ngx_pcalloc failed:%d", NGX_HTTP_INTERNAL_SERVER_ERROR);
        return ; // NGX_HTTP_INTERNAL_SERVER_ERROR;
	}

	if(!g_del_thread_pool){
		del_task_pool_init();
		g_del_thread_pool  = ngx_thread_pool_get((ngx_cycle_t *) ngx_cycle, &name);
	}

	if (g_del_thread_pool == NULL) {
		ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, "get delivery thread pool failed\n");
	}else{
		ret = process_delivery(params, out, 1024, r->connection->log);

		if(ret < 0){
			ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, "%s\n", out);
		}
	}

	r->headers_out.status = NGX_HTTP_OK;
	ngx_str_set(&r->headers_out.content_type, "text/html");

	// allocate a buffer for your response body
	response_body = ngx_pcalloc(r->pool, sizeof(ngx_buf_t));
	if (response_body == NULL) {
		ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, 
						"ngx_pcalloc failed:%d", NGX_HTTP_INTERNAL_SERVER_ERROR);
		return ; // NGX_HTTP_INTERNAL_SERVER_ERROR;
	}

	// send response
	r->headers_out.content_length_n = ngx_strlen(out);
	int rc = ngx_http_send_header(r);
	if (rc == NGX_ERROR || rc > NGX_OK || r->header_only)
	{
		ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, "ngx_http_send_header err:%d", rc);
		return;
	}

	// adjust the pointers of the buffer
	response_body->pos = (u_char *)out;
	response_body->last = (u_char *)out + ngx_strlen(out);
	response_body->memory = 1;    // this buffer is in memory
	response_body->last_buf = 1;  // this is the last buffer in the buffer chain

	ngx_chain_t output;
	output.buf = response_body;
	output.next = NULL;

	r->main->count--;

	ngx_http_output_filter(r,&output);
	
	return;

}

static void
ngx_http_oceanbase_handler_inner(ngx_http_request_t *r)
{
    ngx_http_upstream_t            *u;
    ngx_http_oceanbase_loc_conf_t    *oblcf;
    ngx_str_t                       target;
    ngx_connection_t               *c;
	TXN_TYPE txn_type;
    if (r->subrequest_in_memory) {
        ngx_log_error(NGX_LOG_ALERT, r->connection->log, 0,
                      "ngx_http_oceanbase_module does not support "
                      "subrequest in memory");

        return;
    }

	ngx_http_txn_type(r, &txn_type);

    oblcf = ngx_http_get_module_loc_conf(r, ngx_http_oceanbase_module);

	if(txn_type == DELIVERY){
		ngx_http_oceanbase_handle_delivery(r, &(oblcf->tpcc_delivery_queue));
		return;
	}else if(txn_type == LOGON){
		ngx_http_oceanbase_handle_logon(r);
		return;
	}

    if (ngx_http_upstream_create(r) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, "ngx_http_upstream_create failed\n");
        return;
    }

    u = r->upstream;

    if (oblcf->complex_target) {
        if (ngx_http_complex_value(r, oblcf->complex_target, &target) != NGX_OK)
        {
        	ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, "ngx_http_complex_value failed\n");
            return;
        }

        if (target.len == 0) {
            ngx_log_error(NGX_LOG_ERR, r->connection->log, 0, 
											"oceanbase: handler: empty \"oceanbase_pass\" target");
            return;
        }
    }

    u->conf = &oblcf->upstream;

    u->create_request = ngx_http_oceanbase_create_request;
    u->reinit_request = ngx_http_oceanbase_reinit_request;
    u->process_header = ngx_http_oceanbase_process_header;
    u->abort_request = ngx_http_oceanbase_abort_request;
    u->finalize_request = ngx_http_oceanbase_finalize_request;

    u->input_filter_init = ngx_http_oceanbase_input_filter_init;
    u->input_filter = ngx_http_oceanbase_input_filter;
    u->input_filter_ctx = NULL;

    ngx_http_upstream_oceanbase_init(r);

    /* override the read/write event handler to our own */
    u->write_event_handler = ngx_http_oceanbase_wev_handler;
    u->read_event_handler  = ngx_http_oceanbase_rev_handler;

    /* a bit hack-ish way to return error response (clean-up part) */
    if ((u->peer.connection) && (u->peer.connection->fd == 0)) {
        c = u->peer.connection;
        u->peer.connection = NULL;

        if (c->write->timer_set) {
            ngx_del_timer(c->write);
        }

        ngx_free_connection(c);

        ngx_http_upstream_oceanbase_finalize_request(r, u, NGX_HTTP_SERVICE_UNAVAILABLE);
    }

    return;
}

void
ngx_http_oceanbase_wev_handler(ngx_http_request_t *r, ngx_http_upstream_t *u)
{

	return;
    ngx_connection_t            *c;

    ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                          "oceanbase: upstream write event handler called\n");

    /* just to ensure u->reinit_request always gets called for
     * upstream_next */
    u->request_sent = 1;

    c = u->peer.connection;

    if (ngx_http_upstream_oceanbase_test_connect(c) != NGX_OK) {
        ngx_http_upstream_oceanbase_next(r, u, NGX_HTTP_UPSTREAM_FT_ERROR);
        return;
    }


    (void) ngx_http_oceanbase_process_events(r);
}


void
ngx_http_oceanbase_rev_handler(ngx_http_request_t *r, ngx_http_upstream_t *u)
{
    ngx_connection_t            *c;

    /* just to ensure u->reinit_request always gets called for
     * upstream_next */
    u->request_sent = 1;

    c = u->peer.connection;

    if (ngx_http_upstream_oceanbase_test_connect(c) != NGX_OK) {
        ngx_http_upstream_oceanbase_next(r, u, NGX_HTTP_UPSTREAM_FT_ERROR);
        return;
    }

    (void) ngx_http_oceanbase_process_events(r);
}


static ngx_int_t
ngx_http_oceanbase_create_request(ngx_http_request_t *r)
{
    r->upstream->request_bufs = NULL;

    return NGX_OK;
}


static ngx_int_t
ngx_http_oceanbase_reinit_request(ngx_http_request_t *r)
{
    ngx_http_upstream_t         *u;

    u = r->upstream;

    /* override the read/write event handler to our own */
    u->write_event_handler = ngx_http_oceanbase_wev_handler;
    u->read_event_handler  = ngx_http_oceanbase_rev_handler;

    return NGX_OK;
}


static void
ngx_http_oceanbase_abort_request(ngx_http_request_t *r)
{
}


static void
ngx_http_oceanbase_finalize_request(ngx_http_request_t *r, ngx_int_t rc)
{
}


static ngx_int_t
ngx_http_oceanbase_process_header(ngx_http_request_t *r)
{
    ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                  "ngx_http_oceanbase_process_header should not be called"
                  " by the upstream");

    return NGX_ERROR;
}

static ngx_int_t
ngx_http_oceanbase_input_filter_init(void *data)
{
    ngx_http_request_t          *r = data;

    ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                  "ngx_http_oceanbase_input_filter_init should not be called"
                  " by the upstream");

    return NGX_ERROR;
}

static ngx_int_t
ngx_http_oceanbase_input_filter(void *data, ssize_t bytes)
{
    ngx_http_request_t          *r = data;

    ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                  "ngx_http_oceanbase_input_filter should not be called"
                  " by the upstream");

    return NGX_ERROR;
}

