/*
 * ngx_http_oceanbase_keepalive.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Xiaozhe Wang (chaoslawful)
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#include "ngx_http_oceanbase_keepalive.h"
#include "ngx_http_oceanbase_util.h"
#include "ngx_tpcc_txn.h"
#include <ngx_core.h>
#include <ngx_http.h>
#include <nginx.h>

static void ngx_http_oceanbase_keepalive_dummy_handler(ngx_event_t *ev);
static void ngx_http_oceanbase_keepalive_close_handler(ngx_event_t *ev);

char * ngx_http_upstream_oceanbase_keepalive(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_upstream_oceanbase_srv_conf_t      *oscf = conf;
    ngx_str_t                                   *value;
    ngx_uint_t                                  i;
    ngx_int_t                                   n;
    u_char                                      *data;
    ngx_uint_t                                  len;

    if(oscf->max_cached){
        return "already initialized";
    }

    value = cf->args->elts;

    for (i = 1; i < cf->args->nelts; i++) {
        if (ngx_http_oceanbase_strcmp_const(value[i].data, "max=") == 0) {
            len = value[i].len - (sizeof("max=") - 1);
            data = &value[i].data[sizeof("max=") - 1];

            n = ngx_atoi(data, len);

            if (n == NGX_ERROR || n < 0) {
                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "KeepAlive: invalid \"max\" value \"%V\" in \"%V\" directive",
                                   &value[i], &cmd->name);

                return NGX_CONF_ERROR;
            }

            oscf->max_cached = n;
            continue;
        }

        if (ngx_http_oceanbase_strcmp_const(value[i].data, "overflow=") == 0) {
            len = value[i].len - (sizeof("overflow=") - 1);
            data = &value[i].data[sizeof("overflow=") - 1];

            switch(len){
            case 6:
                if (ngx_http_oceanbase_strcmp_const(data, "reject") == 0) {
                    oscf->overflow = oceanbase_keepalive_overflow_reject;
                } else if (ngx_http_oceanbase_strcmp_const(data, "ignore") == 0) {
                    oscf->overflow = oceanbase_keepalive_overflow_ignore;
                }
                break;

            default:
                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                                   "KeepAlive: invalid \"overflow\" value \"%V\" in \"%V\" directive",
                                   &value[i], &cmd->name);

                return NGX_CONF_ERROR;
            }

            continue;
        }

        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "KeepAlive: invalid parameter \"%V\" in \"%V\" directive",
                           			&value[i], &cmd->name);

        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}


ngx_int_t
ngx_http_oceanbase_keepalive_init(ngx_pool_t *pool, ngx_http_upstream_oceanbase_srv_conf_t *obscf)
{
    ngx_uint_t                              i;
    ngx_http_oceanbase_keepalive_cache_t     *cached;

    cached = ngx_pcalloc(pool, sizeof(ngx_http_oceanbase_keepalive_cache_t) * obscf->max_cached);
    if(cached == NULL){
        return NGX_ERROR;
    }

    ngx_queue_init(&obscf->cache);
    ngx_queue_init(&obscf->free);

    for (i = 0; i < obscf->max_cached; i++) {
        ngx_queue_insert_head(&obscf->free, &cached[i].queue);
        cached[i].srv_conf = obscf;
    }

    return NGX_OK;
}


ngx_int_t
ngx_http_oceanbase_keepalive_get_peer_single(ngx_peer_connection_t *pc,
    										ngx_http_upstream_oceanbase_peer_data_t *obp,
    										ngx_http_upstream_oceanbase_srv_conf_t *obscf)
{
    ngx_http_oceanbase_keepalive_cache_t      *item;
    ngx_queue_t                             *q;
    ngx_connection_t                        *c;

    if (!ngx_queue_empty(&obscf->cache)) {
        q = ngx_queue_head(&obscf->cache);
        ngx_queue_remove(q);

        item = ngx_queue_data(q, ngx_http_oceanbase_keepalive_cache_t, queue);
        c = item->connection;

        ngx_queue_insert_head(&obscf->free, q);

        c->idle = 0;
        c->log = pc->log;
        c->read->log = pc->log;
        c->write->log = pc->log;

        obp->oceanbase_con = item->oceanbase_con;
        obp->used = item->used;
		obp->state = state_db_idle;

        pc->connection = c;
        pc->cached = 1;

        return NGX_DONE;
    }

    return NGX_DECLINED;
}

void
ngx_http_oceanbase_keepalive_free_peer(ngx_peer_connection_t *pc,
    								ngx_http_upstream_oceanbase_peer_data_t *obp,
    								ngx_http_upstream_oceanbase_srv_conf_t *obscf, 
									ngx_uint_t state)
{
    ngx_uint_t                               status;
    ngx_http_oceanbase_keepalive_cache_t      *item;
    ngx_queue_t                             *q;
    ngx_connection_t                        *c;
    ngx_http_upstream_t                     *u;

    if ((state & NGX_PEER_FAILED) 
			|| !obp->oceanbase_con
			|| obp->oceanbase_con->need_reconn){
        obp->failed = 1;
    }

    u = obp->upstream;
    status = u->headers_in.status_n;

    if (!obp->failed
        && pc->connection != NULL
        && (status == NGX_HTTP_NOT_FOUND
            || status == NGX_HTTP_GONE
            || (status == NGX_HTTP_OK && u->header_sent && u->length == 0))){
        c = pc->connection;

        if(ngx_queue_empty(&obscf->free)){
        	ngx_log_error(NGX_LOG_ERR, pc->log, 0, "KeepAlive: free list is empty\n");
            q = ngx_queue_last(&obscf->cache);
            ngx_queue_remove(q);
            item = ngx_queue_data(q, ngx_http_oceanbase_keepalive_cache_t, queue);
			ngx_http_upstream_oceanbase_free_connection(pc->log, item->connection, item->oceanbase_con, obscf);
        }else{
            q = ngx_queue_head(&obscf->free);
            ngx_queue_remove(q);
            item = ngx_queue_data(q, ngx_http_oceanbase_keepalive_cache_t, queue);
        }

        item->connection = c;
        ngx_queue_insert_head(&obscf->cache, q);

        pc->connection = NULL;

        if (c->read->timer_set) {
            ngx_del_timer(c->read);
        }

        if (c->write->timer_set) {
            ngx_del_timer(c->write);
        }

        c->write->handler = ngx_http_oceanbase_keepalive_dummy_handler;
        c->read->handler = ngx_http_oceanbase_keepalive_close_handler;

        c->data = item;
        c->idle = 1;
        c->log = ngx_cycle->log;
        c->read->log = ngx_cycle->log;
        c->write->log = ngx_cycle->log;

		ob_reset_conn(obp->oceanbase_con);

        item->oceanbase_con = obp->oceanbase_con;
        item->used = ++obp->used;
    }
}


static void
ngx_http_oceanbase_keepalive_dummy_handler(ngx_event_t *ev)
{
    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, ev->log, 0, "KeepAlive: dummy handler");
}


static void
ngx_http_oceanbase_keepalive_close_handler(ngx_event_t *ev)
{
    ngx_http_upstream_oceanbase_srv_conf_t    *obscf;
    ngx_http_oceanbase_keepalive_cache_t      *item;

    int                n;
    char               buf[1];
    ngx_connection_t  *c;

    c = ev->data;

    if(c->close){
        goto close;
    }

    n = recv(c->fd, buf, 1, MSG_PEEK);

    if (n == -1 && ngx_socket_errno == NGX_EAGAIN) {
        return;
    }

close:
    item = c->data;
    obscf = item->srv_conf;
    ngx_http_upstream_oceanbase_free_connection(ev->log, c, item->oceanbase_con, obscf);
    ngx_queue_remove(&item->queue);
    ngx_queue_insert_head(&obscf->free, &item->queue);
}
