/*
 * ngx_http_oceanbase_module.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Xiaozhe Wang (chaoslawful)
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#include <assert.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include "ngx_tpcc_common.h"
#include "ngx_http_oceanbase_module.h"
#include "ngx_http_oceanbase_handler.h"
#include "ngx_http_oceanbase_upstream.h"
#include "ngx_http_oceanbase_keepalive.h"
#include "ngx_tpcc_delivery_thread.h"
#include "ngx_tpcc_delivery_pool.h"

/* Forward declaration */

static char *ngx_http_oceanbase_set_complex_value_slot(ngx_conf_t *cf,
    ngx_command_t *cmd, void *conf);
static char *ngx_http_oceanbase_pass(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);
static void *ngx_http_oceanbase_create_loc_conf(ngx_conf_t *cf);
static char *ngx_http_oceanbase_merge_loc_conf(ngx_conf_t *cf, void *parent,
    void *child);
static char *
ngx_http_tpcc_terminal_num(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char *
ngx_http_tpcc_delivery_queue(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char *
ngx_http_tpcc_delivery_runtime(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char *
ngx_http_tpcc_delivery_mdb(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char *
ngx_http_tpcc_max_txn_cnt(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);

//ngx_thread_pool_t * g_del_thread_pool = NULL;

/* config directives for module oceanbase */
static ngx_command_t ngx_http_oceanbase_cmds[] = {
    { ngx_string("oceanbase_server"),
      NGX_HTTP_UPS_CONF|NGX_CONF_1MORE,
      ngx_http_upstream_oceanbase_server,
      NGX_HTTP_SRV_CONF_OFFSET,
      0,
      NULL },

    { ngx_string("oceanbase_keepalive"),
      NGX_HTTP_UPS_CONF|NGX_CONF_1MORE,
      ngx_http_upstream_oceanbase_keepalive,
      NGX_HTTP_SRV_CONF_OFFSET,
      0,
      NULL },

    { ngx_string("oceanbase_dbname"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF
          |NGX_HTTP_LIF_CONF|NGX_CONF_TAKE1,
      ngx_http_oceanbase_set_complex_value_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_oceanbase_loc_conf_t, dsn),
      NULL },

    { ngx_string("oceanbase_pass"),
      NGX_HTTP_LOC_CONF|NGX_HTTP_LIF_CONF|NGX_CONF_TAKE1,
      ngx_http_oceanbase_pass,
      NGX_HTTP_LOC_CONF_OFFSET,
      0,
      NULL },

    { ngx_string("oceanbase_connect_timeout"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF
          |NGX_HTTP_LIF_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_msec_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_oceanbase_loc_conf_t, upstream.connect_timeout),
      NULL },

    { ngx_string("oceanbase_send_query_timeout"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF
          |NGX_HTTP_LIF_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_msec_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_oceanbase_loc_conf_t, upstream.send_timeout),
      NULL },

    { ngx_string("oceanbase_get_result_timeout"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF
          |NGX_HTTP_LIF_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_msec_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_oceanbase_loc_conf_t, get_result_timeout),
      NULL },

	{
		ngx_string("tpcc_terminal_num"),
      	NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
		ngx_http_tpcc_terminal_num,
		NGX_HTTP_LOC_CONF_OFFSET,
      	offsetof(ngx_http_oceanbase_loc_conf_t, tpcc_terminal_num),
		NULL },

	{
		ngx_string("tpcc_delivery_queue"),
      	NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
		ngx_http_tpcc_delivery_queue,
		NGX_HTTP_LOC_CONF_OFFSET,
		0,
		NULL },

	{
		ngx_string("tpcc_delivery_runtime"),
      	NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
		ngx_http_tpcc_delivery_runtime,
		NGX_HTTP_LOC_CONF_OFFSET,
		0,
		NULL },

	{
		ngx_string("tpcc_delivery_mdb"),
      	NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
		ngx_http_tpcc_delivery_mdb,
		NGX_HTTP_LOC_CONF_OFFSET,
		0,
		NULL },

    {
        ngx_string("tpcc_max_txn_cnt"),
        NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
        ngx_http_tpcc_max_txn_cnt,
        NGX_HTTP_LOC_CONF_OFFSET,
        0,
        NULL },

    ngx_null_command
};


/* Nginx HTTP subsystem module hooks */
static ngx_http_module_t ngx_http_oceanbase_module_ctx = {
    NULL,    /* preconfiguration */
    NULL,    /* postconfiguration */

    NULL,    /* create_main_conf */
    NULL,    /* merge_main_conf */

    ngx_http_upstream_oceanbase_create_srv_conf,
             /* create_srv_conf */
    NULL,    /* merge_srv_conf */

    ngx_http_oceanbase_create_loc_conf,    /* create_loc_conf */
    ngx_http_oceanbase_merge_loc_conf      /* merge_loc_conf */
};


ngx_module_t ngx_http_oceanbase_module = {
    NGX_MODULE_V1,
    &ngx_http_oceanbase_module_ctx,       /* module context */
    ngx_http_oceanbase_cmds,              /* module directives */
    NGX_HTTP_MODULE,                    /* module type */
    NULL,    /* init master */
    NULL,    /* init module */
    NULL,    /* init process */
    NULL,    /* init thread */
    NULL,    /* exit thread */
    NULL,    /* exit process */
    NULL,    /* exit master */
    NGX_MODULE_V1_PADDING
};


ngx_oceanbase_http_method_t ngx_oceanbase_http_methods[] = {
    { (u_char *) "GET",       (uint32_t) NGX_HTTP_GET },
    { (u_char *) "HEAD",      (uint32_t) NGX_HTTP_HEAD },
    { (u_char *) "POST",      (uint32_t) NGX_HTTP_POST },
    { (u_char *) "PUT",       (uint32_t) NGX_HTTP_PUT },
    { (u_char *) "DELETE",    (uint32_t) NGX_HTTP_DELETE },
    { (u_char *) "MKCOL",     (uint32_t) NGX_HTTP_MKCOL },
    { (u_char *) "COPY",      (uint32_t) NGX_HTTP_COPY },
    { (u_char *) "MOVE",      (uint32_t) NGX_HTTP_MOVE },
    { (u_char *) "OPTIONS",   (uint32_t) NGX_HTTP_OPTIONS },
    { (u_char *) "PROPFIND" , (uint32_t) NGX_HTTP_PROPFIND },
    { (u_char *) "PROPPATCH", (uint32_t) NGX_HTTP_PROPPATCH },
    { (u_char *) "LOCK",      (uint32_t) NGX_HTTP_LOCK },
    { (u_char *) "UNLOCK",    (uint32_t) NGX_HTTP_UNLOCK },
    { (u_char *) "PATCH",     (uint32_t) NGX_HTTP_PATCH },
    { NULL, 0 }
};


static void *
ngx_http_oceanbase_create_loc_conf(ngx_conf_t *cf)
{
    ngx_http_oceanbase_loc_conf_t             *conf;

    conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_oceanbase_loc_conf_t));
    if (conf == NULL) {
        return NULL;
    }

    conf->upstream.connect_timeout = NGX_CONF_UNSET_MSEC;
    conf->upstream.send_timeout = NGX_CONF_UNSET_MSEC;

    conf->get_result_timeout = NGX_CONF_UNSET_MSEC;

    /* the hardcoded values */
    conf->upstream.cyclic_temp_file = 0;
    conf->upstream.buffering = 0;
    conf->upstream.ignore_client_abort = 0;
    conf->upstream.send_lowat = 0;
    conf->upstream.bufs.num = 0;
    conf->upstream.busy_buffers_size = 0;
    conf->upstream.max_temp_file_size = 0;
    conf->upstream.temp_file_write_size = 0;
    conf->upstream.intercept_errors = 1;
    conf->upstream.intercept_404 = 1;
    conf->upstream.pass_request_headers = 0;
    conf->upstream.pass_request_body = 0;

    conf->complex_target = NGX_CONF_UNSET_PTR;

    conf->tpcc_terminal_num = NGX_CONF_UNSET_SIZE;

    return conf;
}


static char *
ngx_http_oceanbase_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
{
    ngx_http_oceanbase_loc_conf_t *prev = parent;
    ngx_http_oceanbase_loc_conf_t *conf = child;

    ngx_conf_merge_msec_value(conf->upstream.connect_timeout,
                              prev->upstream.connect_timeout, 60000);

    ngx_conf_merge_msec_value(conf->upstream.send_timeout,
                              prev->upstream.send_timeout, 60000);

    ngx_conf_merge_msec_value(conf->get_result_timeout,
                              prev->get_result_timeout, 60000);

    if (conf->dsn == NULL) {
        conf->dsn = prev->dsn;
    }

    ngx_conf_merge_uint_value(conf->tpcc_terminal_num, prev->tpcc_terminal_num,
                              1000000);

    return NGX_CONF_OK;
}


static char *
ngx_http_oceanbase_set_complex_value_slot(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf)
{
    char                             *p = conf;
    ngx_http_complex_value_t        **field;
    ngx_str_t                        *value;
    ngx_http_compile_complex_value_t  ccv;

    field = (ngx_http_complex_value_t **) (p + cmd->offset);

    if (*field) {
        return "is duplicate";
    }

    *field = ngx_palloc(cf->pool, sizeof(ngx_http_complex_value_t));
    if (*field == NULL) {
        return NGX_CONF_ERROR;
    }

    value = cf->args->elts;

    if (value[1].len == 0) {
        ngx_memzero(*field, sizeof(ngx_http_complex_value_t));
        return NGX_OK;
    }

    ngx_memzero(&ccv, sizeof(ngx_http_compile_complex_value_t));

    ccv.cf = cf;
    ccv.value = &value[1];
    ccv.complex_value = *field;

    if (ngx_http_compile_complex_value(&ccv) != NGX_OK) {
        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}

static char *
ngx_http_oceanbase_pass(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_oceanbase_loc_conf_t             *oblcf = conf;
    ngx_http_core_loc_conf_t                *clcf;
    ngx_str_t                               *value;
    ngx_http_compile_complex_value_t         ccv;
    ngx_url_t                                url;
    ngx_uint_t                               n;

    if (oblcf->upstream.upstream) {
        return "is duplicate";
    }

    clcf = ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);

    clcf->handler = ngx_http_oceanbase_handler;

    if (clcf->name.data[clcf->name.len - 1] == '/') {
        clcf->auto_redirect = 1;
    }

    value = cf->args->elts;

    n = ngx_http_script_variables_count(&value[1]);
    if (n) {
        oblcf->complex_target = ngx_palloc(cf->pool,
                                          sizeof(ngx_http_complex_value_t));
        if (oblcf->complex_target == NULL) {
            return NGX_CONF_ERROR;
        }

        ngx_memzero(&ccv, sizeof(ngx_http_compile_complex_value_t));
        ccv.cf = cf;
        ccv.value = &value[1];
        ccv.complex_value = oblcf->complex_target;

        if (ngx_http_compile_complex_value(&ccv) != NGX_OK) {
            return NGX_CONF_ERROR;
        }

        return NGX_CONF_OK;
    }

    oblcf->complex_target = NULL;

    ngx_memzero(&url, sizeof(ngx_url_t));

    url.url = value[1];
    url.no_resolve = 1;

    oblcf->upstream.upstream = ngx_http_upstream_add(cf, &url, 0);

    if (oblcf->upstream.upstream == NULL) {
        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}

TERMINAL_STATES * g_terminal_states = NULL;

static bool init_terminal_status(ngx_conf_t *cf,int terminal_num)
{
	bool bSucc = true;
	if(g_terminal_states){
	}else{
		if(!(g_terminal_states = 
				ngx_palloc(cf->pool, sizeof(TERMINAL_STATES) +
										sizeof(TERMINAL_STATE) * (terminal_num - 1)))){
			bSucc = false;
		}else{
			memset(g_terminal_states, 0, sizeof(TERMINAL_STATES) +
										sizeof(TERMINAL_STATE) * (terminal_num - 1));
			g_terminal_states->max_terminal_cnt = terminal_num;
		}
	}

	return bSucc;
}

static char *
ngx_http_tpcc_terminal_num(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_oceanbase_loc_conf_t* oblcf = conf;
    ngx_conf_set_num_slot(cf, cmd, conf);
    int  terminal_num = oblcf->tpcc_terminal_num;
    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "tpcc_terminal_num: %d", terminal_num);
    bool bSucc = true;

	if(terminal_num <= 0){
		bSucc = false;
	}else if(!(bSucc = init_terminal_status(cf, terminal_num))){
    	ngx_conf_log_error(NGX_LOG_ERR, cf, 0, "init_terminal_status failed terminal_num: %d", terminal_num);
	}else{}

	return bSucc ? NGX_CONF_OK : NGX_CONF_ERROR;
}

static char *
ngx_http_tpcc_delivery_queue(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_str_t   *value = cf->args->elts;
    bool bSucc = true;
	unsigned int i = 0;
	int gid = 0, nid = 0;
	int queue_size = 0;

    for (i = 1; i < cf->args->nelts; i++) {
        if (ngx_strncmp(value[i].data, "group_id=", sizeof("group_id=") - 1) == 0) {
            gid = ngx_atoi(&value[i].data[sizeof("group_id=") - 1], value[i].len - (sizeof("group_id=")- 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "node_id=", sizeof("node_id=") - 1) == 0) {
            nid = ngx_atoi(&value[i].data[sizeof("node_id=") - 1], value[i].len - (sizeof("node_id=") - 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "size=", sizeof("size=") - 1) == 0) {
            queue_size = ngx_atoi(&value[i].data[sizeof("size=") - 1], value[i].len - (sizeof("size=") - 1));
            continue;
        }
	}

    if(queue_size <= 0){
        bSucc = false;
    }else{

		if(queue_size % 32){
			queue_size = (queue_size + 31) / 32 * 32;
		}
		g_del_server.gid = gid;
		g_del_server.nid = nid;
		g_del_server.qsize = queue_size;
	}

    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "tpcc_delivery_queue group_id: %d node_id: %d size:%d", 
												gid,
												nid,
												queue_size);

    return bSucc ? NGX_CONF_OK : NGX_CONF_ERROR;
}

static char *
ngx_http_tpcc_delivery_runtime(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_str_t   *value = cf->args->elts;
    bool bSucc = true;
	unsigned int i = 0;
	
	ngx_uint_t start = 0;
    ngx_uint_t rup = 0;
    ngx_uint_t rdown = 0;
    ngx_uint_t mi = 0;
    ngx_uint_t ri = 0;
    ngx_uint_t wts = 0;

    for (i = 1; i < cf->args->nelts; i++) {
        if (ngx_strncmp(value[i].data, "start=", sizeof("start=") - 1) == 0) {
            start = ngx_atoi(&value[i].data[sizeof("start=") - 1], value[i].len - (sizeof("start=")- 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "rup=", sizeof("rup=") - 1) == 0) {
            rup = ngx_atoi(&value[i].data[sizeof("rup=") - 1], value[i].len - (sizeof("rup=") - 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "rdown=", sizeof("rdown=") - 1) == 0) {
            rdown = ngx_atoi(&value[i].data[sizeof("rdown=") - 1], value[i].len - (sizeof("rdown=") - 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "mi=", sizeof("mi=") - 1) == 0) {
            mi = ngx_atoi(&value[i].data[sizeof("mi=") - 1], value[i].len - (sizeof("mi=") - 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "ri=", sizeof("ri=") - 1) == 0) {
            ri = ngx_atoi(&value[i].data[sizeof("ri=") - 1], value[i].len - (sizeof("ri=") - 1));
            continue;
        }

        if (ngx_strncmp(value[i].data, "wts=", sizeof("wts=") - 1) == 0) {
            wts = ngx_atoi(&value[i].data[sizeof("wts=") - 1], value[i].len - (sizeof("wts=") - 1));
            continue;
        }
	}

    if(start == 0){
        bSucc = false;
    	ngx_conf_log_error(NGX_LOG_ERR, cf, 0, "tpcc_delivery_runtime start: %d rup: %d rdown: %d mi: %d ri: %d wts: %d", 
												start,
												rup,
												rdown,
												mi,
												ri,
												wts);
    }else{
		if(ri <= 1){
			ri = 1;
		}else if(ri <= 5){
			ri = 5;
		}else if(ri <= 10){
			ri = 10;
		}else if(ri <= 15){
			ri = 15;
		}else if(ri <= 20){
			ri = 20;
		}else{
			ri = 30;
		}	

		g_del_runtime.start = start;
		g_del_runtime.rup = rup;
		g_del_runtime.rdown = rdown;
		g_del_runtime.mi = mi;
		g_del_runtime.ri = ri;
		g_del_runtime.wts = wts;
	
    	ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "tpcc_delivery_runtime start: %d rup: %d rdown: %d mi: %d ri: %d wts: %d", 
												start,
												rup,
												rdown,
												mi,
												ri,
												wts);
	}

    return bSucc ? NGX_CONF_OK : NGX_CONF_ERROR;
}

static char *
ngx_http_tpcc_delivery_mdb(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_str_t   *value = cf->args->elts;
    bool bSucc = true;
	unsigned int i = 0;

	char * dsn = (char *)&(g_del_mdb_conn.dsn);
	char * user= (char *)&(g_del_mdb_conn.user);
	char * passwd= (char *)&(g_del_mdb_conn.passwd);
	

    for (i = 1; i < cf->args->nelts; i++) {
        if (ngx_strncmp(value[i].data, "dsn=", sizeof("dsn=") - 1) == 0) {
            strncpy(dsn, (char *)(&value[i].data[sizeof("dsn=") - 1]), value[i].len - (sizeof("dsn=")- 1));
			dsn[value[i].len - (sizeof("dsn=")- 1)] = '\0';
            continue;
        }

        if (ngx_strncmp(value[i].data, "user=", sizeof("user=") - 1) == 0) {
            strncpy(user, (char *)(&value[i].data[sizeof("user=") - 1]), value[i].len - (sizeof("user=") - 1));
			user[value[i].len - (sizeof("user=")- 1)] = '\0';
            continue;
        }

        if (ngx_strncmp(value[i].data, "password=", sizeof("password=") - 1) == 0) {
            strncpy(passwd, (char *)(&value[i].data[sizeof("password=") - 1]), value[i].len - (sizeof("password=") - 1));
			passwd[value[i].len - (sizeof("password=")- 1)] = '\0';
            continue;
        }
	}

    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "tpcc_delivery_mdb dsn: %s user: %s password: %s", 
												dsn,
												user,
												passwd);

    return bSucc ? NGX_CONF_OK : NGX_CONF_ERROR;
}

long long int g_max_txn_cnt = 0;

static char * ngx_http_tpcc_max_txn_cnt(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_oceanbase_loc_conf_t* oblcf = conf;
    ngx_conf_set_num_slot(cf, cmd, conf);
    long long int max_txn_cnt = oblcf->tpcc_max_txn_cnt;
    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "tpcc_max_txn_cnt: %d", max_txn_cnt);
    bool bSucc = true;

    if(max_txn_cnt < 0){
        bSucc = false;
    }else{
        g_max_txn_cnt = max_txn_cnt;
    }

    return bSucc ? NGX_CONF_OK : NGX_CONF_ERROR;
}

