/*
 * ngx_http_oceanbase_output.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#include "ngx_http_oceanbase_module.h"
#include "ngx_http_oceanbase_output.h"
#include "ngx_http_oceanbase_processor.h"
#include "ngx_http_oceanbase_util.h"
#include <nginx.h>

static ngx_int_t ngx_http_oceanbase_get_buf(ngx_http_request_t *r,
    ngx_http_upstream_oceanbase_peer_data_t *obp);
static u_char *ngx_http_oceanbase_get_postponed(ngx_http_request_t *r,
    ngx_http_upstream_oceanbase_peer_data_t *obp, size_t len);

ngx_int_t
ngx_http_oceanbase_output_bufs(ngx_http_request_t *r, ngx_http_upstream_oceanbase_peer_data_t *obp)
{
    ngx_http_upstream_t                    *u = r->upstream;
    ngx_int_t                               rc;
    ngx_chain_t                            *cl;

    if (obp->seen_stream_end) {
        obp->seen_stream_end = 0;

        if (obp->avail_out) {
            cl = ngx_alloc_chain_link(r->pool);
            if (cl == NULL) {
                return NGX_ERROR;
            }

            cl->buf = obp->out_buf;
            cl->next = NULL;
            *obp->last_out = cl;
            obp->last_out = &cl->next;

            obp->avail_out = 0;
        }
    }

    if (!u->header_sent && u->out_bufs) {
        ngx_http_clear_content_length(r);

        r->headers_out.status = NGX_HTTP_OK;

		ngx_str_set(&r->headers_out.content_type, "text/html");

        rc = ngx_http_send_header(r);

        if (rc == NGX_ERROR || rc >= NGX_HTTP_SPECIAL_RESPONSE) {
            return rc;
        }

        u->header_sent = 1;
    }


    for ( ;; ) {
        if (u->out_bufs == NULL) {
            return NGX_OK;
        }

        rc = ngx_http_output_filter(r, u->out_bufs);

        if (rc == NGX_ERROR || rc >= NGX_HTTP_SPECIAL_RESPONSE) {
            return rc;
        }

        ngx_chain_update_chains(r->pool, &u->free_bufs, &u->busy_bufs,
                                &u->out_bufs, u->output.tag);

        obp->last_out = &u->out_bufs;
    }

}

u_char *
ngx_http_oceanbase_request_mem(ngx_http_request_t *r,
    ngx_http_upstream_oceanbase_peer_data_t *obp, size_t len)
{
    ngx_int_t                rc;
    u_char                  *p;

    rc = ngx_http_oceanbase_get_buf(r, obp);

    if (rc != NGX_OK) {
        return NULL;
    }

    if (obp->avail_out < len) {
        p = ngx_http_oceanbase_get_postponed(r, obp, len);
        if (p == NULL) {
            return NULL;
        }

        obp->postponed.pos = p;
        obp->postponed.last = p + len;

        return p;
    }

    return obp->out_buf->last;
}


static ngx_int_t
ngx_http_oceanbase_get_buf(ngx_http_request_t *r,
    ngx_http_upstream_oceanbase_peer_data_t *obp)
{
    ngx_http_upstream_t                 *u = obp->upstream;

    if (obp->avail_out) {
        return NGX_OK;
    }

    if (u->free_bufs) {
        obp->out_buf = u->free_bufs->buf;
        u->free_bufs = u->free_bufs->next;

    } else {
        obp->out_buf = ngx_create_temp_buf(r->pool, 8192);

        if (obp->out_buf == NULL) {
            return NGX_ERROR;
        }

        obp->out_buf->tag = (ngx_buf_tag_t) &ngx_http_oceanbase_module;
        obp->out_buf->recycled = 1;
    }

    obp->avail_out = 8192;

    return NGX_OK;
}


static u_char *
ngx_http_oceanbase_get_postponed(ngx_http_request_t *r,
    ngx_http_upstream_oceanbase_peer_data_t *obp, size_t len)
{
    u_char          *p;

    if (obp->cached.start == NULL) {
        goto alloc;
    }

    if ((size_t) (obp->cached.end - obp->cached.start) < len) {
        ngx_pfree(r->pool, obp->cached.start);
        goto alloc;
    }

    return obp->cached.start;

alloc:

    p = ngx_palloc(r->pool, len);
    if (p == NULL) {
        return NULL;
    }

    obp->cached.start = p;
    obp->cached.end = p + len;

    return p;
}


ngx_int_t
ngx_http_oceanbase_submit_mem(ngx_http_request_t *r,
    ngx_http_upstream_oceanbase_peer_data_t *obp, size_t len)
{
    ngx_chain_t             *cl;
    ngx_int_t                rc;
    size_t                   postponed_len;

    if (obp->postponed.pos != NULL) {
        postponed_len = obp->postponed.last - obp->postponed.pos;

        if (postponed_len > obp->avail_out) {

            if (obp->out_buf && obp->out_buf->pos != obp->out_buf->last) {
                cl = ngx_alloc_chain_link(r->pool);
                if (cl == NULL) {
                    return NGX_ERROR;
                }

                cl->buf = obp->out_buf;
                cl->next = NULL;
                *obp->last_out = cl;
                obp->last_out = &cl->next;
            }

            len = postponed_len > 8192? postponed_len : 8192;

            obp->out_buf = ngx_create_temp_buf(r->pool, len);

            if (obp->out_buf == NULL) {
                return NGX_ERROR;
            }

            obp->out_buf->tag = (ngx_buf_tag_t) &ngx_http_oceanbase_module;
            obp->out_buf->recycled = 1;

            obp->out_buf->last = ngx_copy(obp->out_buf->last, obp->postponed.pos,
                                         postponed_len);

            obp->avail_out = len - postponed_len;

            obp->postponed.pos = NULL;

            if (obp->avail_out == 0) {
                cl = ngx_alloc_chain_link(r->pool);
                if (cl == NULL) {
                    return NGX_ERROR;
                }

                cl->buf = obp->out_buf;
                cl->next = NULL;
                *obp->last_out = cl;
                obp->last_out = &cl->next;
            }

            return NGX_OK;
        }

        for ( ;; ) {
            len = obp->postponed.last - obp->postponed.pos;
            if (len > obp->avail_out) {
                len = obp->avail_out;
            }

            obp->out_buf->last = ngx_copy(obp->out_buf->last, obp->postponed.pos,
                                         len);

            obp->avail_out -= len;

            obp->postponed.pos += len;

            if (obp->postponed.pos == obp->postponed.last) {
                obp->postponed.pos = NULL;
            }

            if (obp->avail_out > 0) {
                break;
            }

            cl = ngx_alloc_chain_link(r->pool);
            if (cl == NULL) {
                return NGX_ERROR;
            }

            cl->buf = obp->out_buf;
            cl->next = NULL;
            *obp->last_out = cl;
            obp->last_out = &cl->next;

            if (obp->postponed.pos == NULL) {
                break;
            }

            rc = ngx_http_oceanbase_get_buf(r, obp);
            if (rc != NGX_OK) {
                return NGX_ERROR;
            }
        }

        return NGX_OK;
    }

    obp->out_buf->last += len;
    obp->avail_out -= len;

    if (obp->avail_out == 0) {

        cl = ngx_alloc_chain_link(r->pool);
        if (cl == NULL) {
            return NGX_ERROR;
        }

        cl->buf = obp->out_buf;
        cl->next = NULL;
        *obp->last_out = cl;
        obp->last_out = &cl->next;
    }

    return NGX_OK;
}
