/*
 * ngx_http_oceanbase_processor.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Xiaozhe Wang (chaoslawful)
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#include <assert.h>
#include "ngx_http_oceanbase_processor.h"
#include "ngx_http_oceanbase_module.h"
#include "ngx_http_oceanbase_util.h"
#include "ngx_http_oceanbase_output.h"
#include "ngx_http_oceanbase_upstream.h"

static ngx_int_t ngx_http_upstream_oceanbase_connect(ngx_http_request_t *r,
    ngx_connection_t *c, ngx_http_upstream_oceanbase_peer_data_t *obp,
    oceanbase_connection *obc);
static ngx_int_t ngx_http_upstream_oceanbase_send_query(ngx_http_request_t *r,
    ngx_connection_t *c, ngx_http_upstream_oceanbase_peer_data_t *obp,
    oceanbase_connection *obc);
static ngx_int_t ngx_http_upstream_oceanbase_get_result(ngx_http_request_t *r,
    ngx_connection_t *c, ngx_http_upstream_oceanbase_peer_data_t *obp,
    oceanbase_connection *obc);

ngx_int_t
ngx_http_oceanbase_process_events(ngx_http_request_t *r)
{
    ngx_http_upstream_t                         *u;
    ngx_connection_t                            *c;
    ngx_http_upstream_oceanbase_peer_data_t     *obp;
    oceanbase_connection                            *obc;
    ngx_int_t                                   rc;

    u = r->upstream;
    c = u->peer.connection;

    obp = u->peer.data;
    obc = obp->oceanbase_con;

    switch (obp->state) {
    case state_db_connect:
        rc = ngx_http_upstream_oceanbase_connect(r, c, obp, obc);
        break;

    case state_db_idle: 
        c->log->action = "sending query to oceanbase upstream";

    case state_db_send_query:
        rc = ngx_http_upstream_oceanbase_send_query(r, c, obp, obc);
        break;

	case state_db_get_result:
        c->log->action = "get result from oceanbase upstream";
		rc = ngx_http_upstream_oceanbase_get_result(r, c, obp, obc);
		break;

    default:
        ngx_log_error(NGX_LOG_ERR, c->log, 0,
                      "unknown state: %d", (int) obp->state);
        return NGX_ERROR;
    }

    if (rc == NGX_ERROR) {
        ngx_http_upstream_oceanbase_next(r, u, NGX_HTTP_UPSTREAM_FT_ERROR);
        return NGX_ERROR;
    }

    if (rc >= NGX_HTTP_SPECIAL_RESPONSE) {
        ngx_http_upstream_oceanbase_finalize_request(r, u, rc);
        return NGX_ERROR;
    }

    return rc;
}


static ngx_int_t
ngx_http_upstream_oceanbase_connect(ngx_http_request_t *r,
    ngx_connection_t *c, ngx_http_upstream_oceanbase_peer_data_t *obp,
    oceanbase_connection *obc)
{
    oceanbase_return_t             ret;

    ret = ob_connect(obc);
    if (c->write->timer_set){
        ngx_del_timer(c->write);
    }

    if (ret != OCEANBASE_RETURN_OK){
        ngx_log_error(NGX_LOG_ERR, c->log, 0, "failed to connect OceanBase or prepare statements: %d: %s",
                      (int) ret, oceanbase_error_msg(obc));

        return NGX_ERROR;
    }

    c->log->action = "sending query to OceanBase database";

    return ngx_http_upstream_oceanbase_send_query(r, c, obp, obc);
}


int ngx_http_txn_type(ngx_http_request_t *r, TXN_TYPE *txn_type)
{
    int ret = 0;

    switch(r->uri.len){
    case    sizeof(NEW_ORDER_URI) - 1: //new_order
        if(!ngx_strncmp(r->uri.data, NEW_ORDER_URI, sizeof(NEW_ORDER_URI) - 1)){
            *txn_type = NEW_ORDER;
        }else{}
        break;
    case    sizeof(PAYMENT_URI) - 1:  //payment
        if(!ngx_strncmp(r->uri.data, PAYMENT_URI, sizeof(PAYMENT_URI) - 1)){
            *txn_type = PAYMENT;
        }else{}
        break;
    case    sizeof(DELIVERY_URI) - 1:  //delivery
        if(!ngx_strncmp(r->uri.data, DELIVERY_URI, sizeof(DELIVERY_URI) - 1)){
            *txn_type = DELIVERY;
        }else{}
        break;
    case    sizeof(ORDER_STATUS_URI) - 1: //order_status
        if(!ngx_strncmp(r->uri.data, ORDER_STATUS_URI, sizeof(ORDER_STATUS_URI) - 1)){
            *txn_type = ORDER_STATUS;
        }else{}
       break;
    case    sizeof(STOCK_LEVEL_URI) - 1: //stock_level
        if(!ngx_strncmp(r->uri.data, STOCK_LEVEL_URI, sizeof(STOCK_LEVEL_URI) - 1)){
            *txn_type = STOCK_LEVEL;
        }else{}
        break;
    case    sizeof(LOGON_URI) - 1: //logon
        if(!ngx_strncmp(r->uri.data, LOGON_URI, sizeof(LOGON_URI) - 1)){
            *txn_type = LOGON;
        }else{}
        break;
    default:
        *txn_type =  TXN_UNKNOWN;
        break;
    }

    if(*txn_type == TXN_UNKNOWN){
        ret = -1;
    }else{}

    return ret;
}

extern int process_tpcc_txns(oceanbase_connection *obc, ngx_log_t *log);
extern int oceanbase_get_result(oceanbase_connection * ob_st, ngx_log_t *log);
extern ngx_int_t ngx_http_oceanbase_output_tpcc_txns(ngx_http_request_t *r, oceanbase_connection *obc, bool succ);

static ngx_int_t
ngx_http_upstream_oceanbase_send_query(ngx_http_request_t *r,
    ngx_connection_t *c, ngx_http_upstream_oceanbase_peer_data_t *obp,
    oceanbase_connection *obc)
{
    oceanbase_return_t           ret = OCEANBASE_RETURN_OK;
	TXN_TYPE txn_type; 

	if(!ngx_http_txn_type(r, &txn_type)){
	 	ngx_chain_t *bufs = r->request_body->bufs;
		int len = 0;

		while(bufs){
			 ngx_buf_t *buf = bufs->buf;
			 memcpy(obc->params + len, buf->pos, buf->last - buf->pos);
			 len += (buf->last - buf->pos);
			 bufs = bufs->next;
		}

		obc->params[len] = '\0';
		obc->txn_type = txn_type;
		ret = process_tpcc_txns(obc, c->log);
	}else{
		assert(false);
	}

    if (ret == OCEANBASE_RETURN_IO_WAIT) {
        if (obp->state != state_db_get_result) {
            obp->state = state_db_get_result;

            if (c->write->timer_set) {
                ngx_del_timer(c->write);
            }
        }
        return NGX_AGAIN;
    }

    if (c->write->timer_set) {
        ngx_del_timer(c->write);
    }

    if (ret != OCEANBASE_RETURN_OK) {
        ngx_log_error(NGX_LOG_ERR, c->log, 0,
                      "process_tpcc_txns(txn_type:%d params:%s) failed, %s",
						obc->txn_type, obc->params, obc->msg);
		ngx_http_oceanbase_output_tpcc_txns(r, obc, false);
        ngx_http_upstream_oceanbase_done(r, r->upstream, obp, NGX_DONE);

        return NGX_DONE;
    }else{
        ngx_log_error(NGX_LOG_WARN, c->log, 0,
                      "process_tpcc_txns(txn_type:%d params:%s) returns OCEANBASE_RETURN_OK when called the first time",
                      obc->txn_type, obc->params);
		ngx_http_oceanbase_output_tpcc_txns(r, obc, true);
        ngx_http_upstream_oceanbase_done(r, r->upstream, obp, NGX_DONE);
        return NGX_DONE;
	}
}


ngx_int_t
ngx_http_upstream_oceanbase_get_result(ngx_http_request_t *r,
    ngx_connection_t *c, ngx_http_upstream_oceanbase_peer_data_t *obp,
    oceanbase_connection *obc)
{
	int ret = 0;
	ngx_http_upstream_t             *u = r->upstream;
	
	ret = oceanbase_get_result(obc, c->log);	

	if(ret == OCEANBASE_RETURN_IO_WAIT){
		ret = NGX_AGAIN;
	}else{
		ret = ngx_http_oceanbase_output_tpcc_txns(r, obc, ret == OCEANBASE_RETURN_OK);

		if(ret < 0){
			ngx_log_error(NGX_LOG_ERR, c->log, 0,
                      "ngx_http_upstream_oceanbase_get_result(%d) returns error %d",
                      obc->txn_type, ret);
		}
		ngx_http_upstream_oceanbase_done(r, u, obp, NGX_DONE);
        ret = NGX_DONE;
	}
	
	return ret;
}


void
ngx_http_upstream_oceanbase_done(ngx_http_request_t *r,
    ngx_http_upstream_t *u, ngx_http_upstream_oceanbase_peer_data_t *obp,
    ngx_int_t rc)
{
    ngx_connection_t            *c;

    (void) ngx_http_oceanbase_output_bufs(r, obp);

    u->length = 0;

    if (rc == NGX_DONE) {
        u->header_sent = 1;
        u->headers_in.status_n = NGX_HTTP_OK;
        rc = NGX_OK;

    } else {
        r->headers_out.status = rc;
        u->headers_in.status_n = rc;
    }

    c = u->peer.connection;

    c->log->action = "being idle";

    obp->state = state_db_idle;

    ngx_http_upstream_oceanbase_finalize_request(r, u, rc);
}
