/* 
 * ngx_http_oceanbase_upstream.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#include "ngx_http_oceanbase_module.h"
#include "ngx_http_oceanbase_upstream.h"
#include "ngx_http_oceanbase_keepalive.h"
#include "ngx_http_oceanbase_processor.h"
#include "ngx_http_oceanbase_util.h"
#include "ngx_tpcc_txn.h"
#include "ngx_tpcc_delivery_thread.h"

static ngx_int_t ngx_http_upstream_oceanbase_init_ups(ngx_conf_t *cf,
        ngx_http_upstream_srv_conf_t *uscf);

static ngx_int_t ngx_http_upstream_oceanbase_init_peer(ngx_http_request_t *r,
    ngx_http_upstream_srv_conf_t *uscf);

static ngx_int_t ngx_http_upstream_oceanbase_get_peer(ngx_peer_connection_t *pc,
        void *data);

static void ngx_http_upstream_oceanbase_free_peer(ngx_peer_connection_t *pc,
        void *data, ngx_uint_t state);

static ngx_int_t ngx_http_oceanbase_output_filter(void *data, ngx_chain_t *in);

void *
ngx_http_upstream_oceanbase_create_srv_conf(ngx_conf_t *cf)
{
    ngx_http_upstream_oceanbase_srv_conf_t    *conf;

    conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_upstream_oceanbase_srv_conf_t));
    if(conf == NULL){
        return NULL;
    }

    conf->pool = cf->pool;
    return conf;
}


char *
ngx_http_upstream_oceanbase_server(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_upstream_oceanbase_srv_conf_t        *obscf = conf;
    ngx_http_upstream_oceanbase_server_t          *obs;
    ngx_str_t                                   *value;
    ngx_uint_t                                   i;
    ngx_http_upstream_srv_conf_t                *uscf;

    uscf = ngx_http_conf_get_module_srv_conf(cf, ngx_http_upstream_module);

    if(obscf->servers == NULL) {
        obscf->servers = ngx_array_create(cf->pool, 4,
                                 sizeof(ngx_http_upstream_oceanbase_server_t));
        if (obscf->servers == NULL) {
            return NGX_CONF_ERROR;
        }

        uscf->servers = obscf->servers;
    }

    if(!(obs = ngx_array_push(obscf->servers))){
        return NGX_CONF_ERROR;
    }

    ngx_memzero(obs, sizeof(ngx_http_upstream_oceanbase_server_t));

    value = cf->args->elts;

    for (i = 1; i < cf->args->nelts; i++) {
        if (ngx_strncmp(value[i].data, "dsn=", sizeof("dsn=") - 1) == 0) {
            obs->dsn.len = value[i].len - (sizeof("dsn=") - 1);
            obs->dsn.data = &value[i].data[sizeof("dsn=") - 1];
            continue;
        }

        if (ngx_strncmp(value[i].data, "user=", sizeof("user=") - 1) == 0) {
            obs->user.len = value[i].len - (sizeof("user=") - 1);
            obs->user.data = &value[i].data[sizeof("user=") - 1];
            continue;
        }

        if (ngx_strncmp(value[i].data, "password=", sizeof("password=") - 1) == 0) {
            obs->password.len = value[i].len - (sizeof("password=") - 1);
            obs->password.data = &value[i].data[sizeof("password=") - 1];
            continue;
        }
		
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "invalid parameter \"%V\" in oceanbase_server", &value[i]);

        return NGX_CONF_ERROR;
    }

	strncpy(g_del_conn.dsn, (char *)obs->dsn.data, obs->dsn.len);
	g_del_conn.dsn[obs->dsn.len] = '\0';	

	strncpy(g_del_conn.user, (char *)obs->user.data, obs->user.len);
	g_del_conn.user[obs->user.len] = '\0';	

	strncpy(g_del_conn.passwd, (char *)obs->password.data, obs->password.len);
	g_del_conn.passwd[obs->password.len] = '\0';	

    uscf->peer.init_upstream = ngx_http_upstream_oceanbase_init_ups;

    return NGX_CONF_OK;
}


static ngx_int_t
ngx_http_upstream_oceanbase_init_ups(ngx_conf_t *cf, ngx_http_upstream_srv_conf_t *uscf)
{
    ngx_http_upstream_oceanbase_srv_conf_t    *obscf;
    ngx_http_upstream_oceanbase_server_t      *server;
    ngx_http_upstream_oceanbase_peer_t       *peer;

    uscf->peer.init = ngx_http_upstream_oceanbase_init_peer;

    obscf = ngx_http_conf_upstream_srv_conf(uscf, ngx_http_oceanbase_module);

    if(!obscf->servers || obscf->servers->nelts == 0){
        ngx_log_error(NGX_LOG_EMERG, cf->log, 0, "no oceanbase_server defined in upstream");

        return NGX_ERROR;
    }

    server = uscf->servers->elts;

    peer = ngx_pcalloc(cf->pool, sizeof(ngx_http_upstream_oceanbase_peer_t));

    if(peer == NULL){
        return NGX_ERROR;
    }

	peer->dsn = server[0].dsn;
	peer->user= server[0].user;
	peer->password = server[0].password;	

    obscf->peer = peer;
    obscf->active_conns = 0;

    if(obscf->max_cached){
        return ngx_http_oceanbase_keepalive_init(cf->pool, obscf);
    }

    return NGX_OK;
}

static ngx_int_t
ngx_http_upstream_oceanbase_init_peer(ngx_http_request_t *r, ngx_http_upstream_srv_conf_t *uscf)
{
    ngx_http_upstream_oceanbase_peer_data_t   *obp;
    ngx_http_upstream_oceanbase_srv_conf_t    *obscf;
    ngx_http_upstream_t                     *u;
    ngx_http_oceanbase_loc_conf_t             *oblcf;
    ngx_str_t                                dsn;

    obp = ngx_pcalloc(r->pool, sizeof(ngx_http_upstream_oceanbase_peer_data_t));
    if (obp == NULL) {
        goto failed;
    }

    u = r->upstream;

    obp->upstream = u;
    obp->request  = r;
    obp->last_out = &u->out_bufs;

    obscf = ngx_http_conf_upstream_srv_conf(uscf, ngx_http_oceanbase_module);
    obp->srv_conf = obscf;

    oblcf = ngx_http_get_module_loc_conf(r, ngx_http_oceanbase_module);
    obp->loc_conf = oblcf;

    obp->dsn.len = 0;

    u->output.output_filter = ngx_http_oceanbase_output_filter;
    u->output.filter_ctx = r;
    u->output.in   = NULL;
    u->output.busy = NULL;

    u->peer.data = obp;
    u->peer.get = ngx_http_upstream_oceanbase_get_peer;
    u->peer.free = ngx_http_upstream_oceanbase_free_peer;

    if(oblcf->dsn){
        if(ngx_http_complex_value(r, oblcf->dsn, &dsn) != NGX_OK){
            goto failed;
        }
        if(dsn.len){
            obp->dsn = dsn;
        }
    }

    return NGX_OK;

failed:

    return NGX_ERROR;
}


static ngx_int_t
ngx_http_upstream_oceanbase_get_peer(ngx_peer_connection_t *pc, void *data)
{
    ngx_http_upstream_oceanbase_peer_data_t   *obp = data;
    ngx_http_upstream_oceanbase_srv_conf_t    *obscf;
    ngx_http_upstream_oceanbase_peer_t       *peer;
    ngx_connection_t                        *c = NULL;
    oceanbase_connection                        *obc = NULL;
    ngx_str_t                                dsn;
    oceanbase_return_t                       ret;
    ngx_socket_t                             fd;
    ngx_event_t                             *rev, *wev;
    ngx_int_t                                rc;
    ngx_int_t                                event;

    obscf = obp->srv_conf;

    obp->failed = 0;

    if(obscf->max_cached){
        rc = ngx_http_oceanbase_keepalive_get_peer_single(pc, obp, obscf);
        if(rc != NGX_DECLINED){
            return rc;
        }
    }

    peer = obscf->peer;

    if (obscf->overflow == oceanbase_keepalive_overflow_reject
        && obscf->active_conns >= obscf->max_cached)
    {
        ngx_log_error(NGX_LOG_ERR, pc->log, 0,
                      "oceanbase: connection pool full, rejecting request "
	                      "to upstream");

        pc->connection = ngx_get_connection(0, pc->log);

        return NGX_AGAIN;
    }

    obc = ngx_pcalloc(obscf->pool, sizeof(oceanbase_connection));

    if(obc == NULL){
        return NGX_ERROR;
    }

    obp->oceanbase_con = obc;

    if(obp->dsn.len){
        dsn = obp->dsn;
    }else{
        dsn = peer->dsn;
    }

    ngx_memcpy(obc->dsn, dsn.data, dsn.len);
    obc->dsn[dsn.len] = '\0';

    ngx_memcpy(obc->user, peer->user.data, peer->user.len);
    obc->user[peer->user.len] = '\0';

    ngx_memcpy(obc->password, peer->password.data, peer->password.len);
    obc->password[peer->password.len] = '\0';

    ret = ob_connect(obc);

    if(ret != OCEANBASE_RETURN_OK) {
        ngx_log_error(NGX_LOG_EMERG, pc->log, 0,
                      "failed to connect OceanBase: %d: %s",
                      (int) ret,
                      oceanbase_error_msg(obc));

        ob_free_conn(obc);
        ngx_pfree(obscf->pool, obc);

        return NGX_DECLINED;
    }

    obscf->active_conns++;

    fd = ob_get_conn_fd(obc);

    if(fd == -1){
        ngx_log_error(NGX_LOG_ERR, pc->log, 0,
                      "failed to get OceanBase connection fd");
        goto invalid;
    }

    c = ngx_get_connection(fd, pc->log);
    if (c == NULL) {
        ngx_log_error(NGX_LOG_ERR, pc->log, 0,
                      "failed to get a free nginx connection");

        goto invalid;
    }

    c->log = pc->log;
    c->log_error = pc->log_error;
    c->number = ngx_atomic_fetch_add(ngx_connection_counter, 1);

    rev = c->read;
    wev = c->write;

    rev->log = pc->log;
    wev->log = pc->log;

    pc->connection = c;

	obp->state = state_db_idle;

    if(ngx_add_conn){
        if(ngx_add_conn(c) == NGX_ERROR){
            goto invalid;
        }

        wev->ready = 1;
        return NGX_DONE;
    }

    if (ngx_event_flags & NGX_USE_CLEAR_EVENT) {
        event = NGX_CLEAR_EVENT;
    } else {
        event = NGX_LEVEL_EVENT;
    }

    if (ngx_add_event(rev, NGX_READ_EVENT, event) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, pc->log, 0, "failed to add nginx event");

        goto invalid;
    }

    wev->ready = 1;

    return NGX_DONE;

invalid:

    ngx_http_upstream_oceanbase_free_connection(pc->log, pc->connection, obc, obscf);

    return NGX_ERROR;
}


static void
ngx_http_upstream_oceanbase_free_peer(ngx_peer_connection_t *pc, void *data, ngx_uint_t state)
{
    ngx_http_upstream_oceanbase_peer_data_t   *obp = data;
    ngx_http_upstream_oceanbase_srv_conf_t    *obscf;

    obscf = obp->srv_conf;

    if(obscf->max_cached) {
        ngx_http_oceanbase_keepalive_free_peer(pc, obp, obscf, state);
    }

    if (pc && pc->connection) {
        ngx_http_upstream_oceanbase_free_connection(pc->log, pc->connection,
                                                  obp->oceanbase_con, obscf);
        obp->oceanbase_con = NULL;
        pc->connection = NULL;
    }
}


static ngx_int_t
ngx_http_oceanbase_output_filter(void *data, ngx_chain_t *in)
{
    ngx_http_request_t              *r = data;

    r->upstream->request_sent = 1;

    (void) ngx_http_oceanbase_process_events(r);

    return NGX_AGAIN;
}

void
ngx_http_upstream_oceanbase_free_connection(ngx_log_t *log,
    ngx_connection_t *c, oceanbase_connection *obc,
    ngx_http_upstream_oceanbase_srv_conf_t *obscf)
{
    ngx_event_t  *rev, *wev;
	
	ngx_log_error(NGX_LOG_ERR, log, 0, "upstream free oceanbase connection");

    obscf->active_conns--;

    if(obc){
        ob_free_conn(obc);
        ngx_pfree(obscf->pool, obc);
    }

    if(c){
        rev = c->read;
        wev = c->write;

        if(rev->timer_set){
            ngx_del_timer(rev);
        }

        if(wev->timer_set){
            ngx_del_timer(wev);
        }

        if(ngx_del_conn){
            ngx_del_conn(c, NGX_CLOSE_EVENT);
        }else{
            if(rev->active || rev->disabled){
                ngx_del_event(rev, NGX_READ_EVENT, NGX_CLOSE_EVENT);
            }

            if(wev->active || wev->disabled){
                ngx_del_event(wev, NGX_WRITE_EVENT, NGX_CLOSE_EVENT);
            }
        }

        if(rev->posted){
            ngx_delete_posted_event(rev);
        }

        if(wev->posted){
            ngx_delete_posted_event(wev);
        }

        rev->closed = 1;
        wev->closed = 1;

        ngx_free_connection(c);

        c->fd = (ngx_socket_t) -1;
    }
}

