/*
 * ngx_tpcc_delivery_pool.c
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */
 
#include <assert.h>
#include <ngx_thread.h>
#include "ngx_tpcc_common.h"
#include "ngx_tpcc_delivery_pool.h"
#include "ngx_tpcc_delivery_thread.h"

#define BITS_PER_UINT32 (8* sizeof(uint32_t))

static ngx_thread_task_pool_t * g_del_task_pool = NULL;

ngx_int_t del_task_pool_init() 
{
    if (g_del_task_pool != NULL) 
		return NGX_OK;

    ngx_log_t* log = ((ngx_cycle_t *) ngx_cycle)->log;
    ngx_pool_t *pool = ((ngx_cycle_t *) ngx_cycle)->pool;

	DeliveryTask 		* del_task = NULL;
    ngx_thread_task_t   * task = NULL;

    g_del_task_pool = ngx_pcalloc(pool, sizeof(ngx_thread_task_pool_t));
    if (g_del_task_pool == NULL) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "ngx_thread_task_pool_t calloc failed.");
        return NGX_ERROR;
    }

    if (ngx_thread_mutex_create(&(g_del_task_pool->mtx), log) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "ngx_thread_mutex_create failed.");
        return NGX_ERROR;
    }

    ngx_int_t pool_size = g_del_server.qsize;
	assert(pool_size % 32 == 0); // for easy use
    g_del_task_pool->pool_size = pool_size;
    g_del_task_pool->tasks = ngx_pcalloc(pool, sizeof(ngx_thread_task_t*) * pool_size);
    if (g_del_task_pool->tasks  == NULL) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "ngx_pcalloc tasks failed.");
        return NGX_ERROR;
    }

    g_del_task_pool->used_flag = ngx_pcalloc(pool, sizeof(uint32_t) * (pool_size / BITS_PER_UINT32));
    if (g_del_task_pool->used_flag  == NULL) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "ngx_pcalloc used_flag failed.");
        return NGX_ERROR;
    }

    ngx_int_t i = 0;
    for (i = 0; i < pool_size; i++) {
        task = ngx_thread_task_alloc(pool, sizeof(DeliveryTask));

        if (task == NULL) {
            return NGX_ERROR;
        }

        del_task = (DeliveryTask *)(task->ctx);
        del_task->pos = i;
        g_del_task_pool->tasks[i] = task;
    }
    return NGX_OK;
}

ngx_thread_task_t* get_del_task_from_pool() 
{
    ngx_log_t* log = ((ngx_cycle_t *) ngx_cycle)->log;
	int pos = 0;

    if (ngx_thread_mutex_lock(&(g_del_task_pool->mtx), log) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "lock failed found");
        return NULL;
    }

    ngx_uint_t i = 0;
    ngx_thread_task_t* task = NULL;

    for (i = 0; i < g_del_task_pool->pool_size / BITS_PER_UINT32; i++) {
        if (g_del_task_pool->used_flag[i] == 0) {
            task = g_del_task_pool->tasks[i * BITS_PER_UINT32];
            g_del_task_pool->used_flag[i] |= 1;
            break;
        } else if (g_del_task_pool->used_flag[i] == UINT32_MAX){
            continue;
        } else{
			pos = __builtin_ffs(~(g_del_task_pool->used_flag[i]));

			if(pos == 0){
				continue;
			}else{
				task = g_del_task_pool->tasks[i * BITS_PER_UINT32 + pos - 1];
				g_del_task_pool->used_flag[i] |= 1 << (pos - 1);	
			}
			break;
		}
    }

	if(!task){
		char msg[256] = {'\0'};

		sprintf(msg, "gets: %lld puts: %lld", g_del_task_pool->gets, g_del_task_pool->puts);
        ngx_log_error(NGX_LOG_ERR, log, 0, "no empty task found in pool pool_size: %d %s\n", 
										g_del_task_pool->pool_size, msg);
	}else{
		g_del_task_pool->gets ++;
        ngx_log_error(NGX_LOG_INFO, log, 0, "get task Task pos: %d UsedFlag: i: %d off: %d\n", 
										((DeliveryTask*)(task->ctx))->pos, i, pos);
	}

    if (ngx_thread_mutex_unlock(&(g_del_task_pool->mtx), log) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "lock failed found");
    }

    return task;
}

ngx_int_t return_task_to_pool(DeliveryTask* del_task) 
{
    ngx_log_t* log = ((ngx_cycle_t *) ngx_cycle)->log;

    if (ngx_thread_mutex_lock(&(g_del_task_pool->mtx), log) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "lock failed found");
        return NGX_ERROR;
    }

    ngx_int_t pos = del_task->pos;
    g_del_task_pool->used_flag[pos / BITS_PER_UINT32] &= ~(1 << (pos % BITS_PER_UINT32) );
    ngx_thread_task_t* task = g_del_task_pool->tasks[pos];
    ngx_memzero(task, sizeof(*task));
    task->ctx = del_task;

	g_del_task_pool->puts ++;

    ngx_log_error(NGX_LOG_INFO, log, 0, "return task: pos: %d \n", pos);
		
    if (ngx_thread_mutex_unlock(&(g_del_task_pool->mtx), log) != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, log, 0, "lock failed found");
        return NGX_ERROR;
    }
    return NGX_OK;
}

void
del_thread_completion(ngx_event_t *ev)
{
    DeliveryTask* del_task = ev->data;
    return_task_to_pool(del_task);
    /* executed in nginx event loop */
}
